from functools import partial
from contextlib import suppress
from typing import Any, Awaitable, Callable, Optional, Union, AsyncGenerator
import inspect

from .datatypes import CoroutineFunction
from .compat import wrap_async, wrap_generator_async, AsyncExitStack
from .exceptions import ProviderDeclarationError
from . import scopes


async def _terminate_agen(async_gen: AsyncGenerator):
    with suppress(StopAsyncIteration):
        await async_gen.asend(None)


class Provider:
    """Base class for providers.

    This is mostly a wrapper around a provider function, along with
    some metadata.
    """

    __slots__ = ("func", "name", "scope", "lazy")

    def __init__(self, func: Callable, name: str, scope: str, lazy: bool):
        if lazy and scope != scopes.FUNCTION:
            raise ProviderDeclarationError(
                "Lazy providers must be function-scoped"
            )

        if inspect.isgeneratorfunction(func):
            func = wrap_generator_async(func)
        elif inspect.isasyncgenfunction(func):
            pass
        elif not inspect.iscoroutinefunction(func):
            func = wrap_async(func)

        assert inspect.iscoroutinefunction(func) or inspect.isasyncgenfunction(
            func
        )

        self.func: Union[AsyncGenerator, CoroutineFunction] = func
        self.name = name
        self.scope = scope
        self.lazy = lazy

    @classmethod
    def create(cls, func, **kwargs) -> "Provider":
        """Factory method to build a provider of the appropriate scope."""
        scope: Optional[str] = kwargs.get("scope")
        if scope == scopes.SESSION:
            return SessionProvider(func, **kwargs)
        return FunctionProvider(func, **kwargs)

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} "
            "name={self.name}, scope={self.scope}, func={self.func}>"
        )

    # NOTE: the returned value is an awaitable, so we *must not*
    # declare this function as `async` — its return value should already be.
    def __call__(self, stack: AsyncExitStack) -> Awaitable:
        raise NotImplementedError


class FunctionProvider(Provider):
    """Represents a function-scoped provider.

    Its value is recomputed every time the provider is called.
    """

    def __call__(self, stack: AsyncExitStack) -> Awaitable:
        value: Union[Awaitable, AsyncGenerator] = self.func()

        if inspect.isasyncgen(value):
            agen = value
            # We cannot use `await` in here => define a coroutine function
            # and return the (awaitable) coroutine itself.

            async def get_value() -> Any:
                # Executes setup + `yield <some_value>`.
                val = await agen.asend(None)
                # Registers cleanup to be executed when the stack exits.
                stack.push_async_callback(partial(_terminate_agen, agen))
                return val

            value: Awaitable = get_value()

        return value


class SessionProvider(Provider):
    """Represents a session-scoped provider.

    When called, it builds its instance if necessary and returns it. This
    means that the underlying provider is only built once and is reused
    across function calls.
    """

    __slots__ = Provider.__slots__ + ("_instance", "_generator")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._instance: Optional[Any] = None
        self._generator: Optional[AsyncGenerator] = None

    async def enter_session(self):
        value: Union[Awaitable, AsyncGenerator] = self.func()

        if inspect.isasyncgen(value):
            agen = value
            value = await agen.asend(None)
            self._generator = agen
        else:
            value = await value

        self._instance = value

    async def exit_session(self):
        if self._generator is not None:
            await _terminate_agen(self._generator)
            self._generator = None
        self._instance = None

    async def _get_instance(self) -> Any:
        if self._instance is None:
            await self.enter_session()
        return self._instance

    def __call__(self, stack: AsyncExitStack) -> Awaitable:
        return self._get_instance()



import inspect
from contextlib import contextmanager, suppress
from functools import partial, wraps
from importlib import import_module
from typing import Callable, Dict, List, NamedTuple, Optional, Tuple, Union

from . import scopes
from .compat import AsyncExitStack, wrap_async
from .datatypes import CoroutineFunction
from .exceptions import (
    ConsumerDeclarationError,
    RecursiveProviderError,
    UnknownScope,
)
from .providers import Provider, SessionProvider

PositionalProviders = List[Tuple[str, Provider]]
KeywordProviders = Dict[str, Provider]

# Sentinel for parameters that have no provider.
_NO_PROVIDER = object()


class ResolvedProviders(NamedTuple):

    positional: PositionalProviders
    keyword: KeywordProviders

    def __bool__(self):
        return bool(self.positional) or bool(self.keyword)


class Store:

    __slots__ = (
        "providers",
        "scope_aliases",
        "default_scope",
        "providers_module",
        "_session_providers",
    )

    def __init__(
        self,
        providers_module="providerconf",
        scope_aliases: Dict[str, str] = None,
        default_scope: str = scopes.FUNCTION,
    ):
        if scope_aliases is None:
            scope_aliases = {}

        self.providers: Dict[str, Provider] = {}
        self._session_providers: Dict[str, SessionProvider] = {}
        self.scope_aliases = scope_aliases
        self.default_scope = default_scope
        self.providers_module = providers_module

    def empty(self):
        return not self.providers

    def has_provider(self, name: str) -> bool:
        return name in self.providers

    def _get(self, name: str) -> Optional[Provider]:
        return self.providers.get(name)

    def discover_default(self):
        with suppress(ImportError):
            self.discover(self.providers_module)

    def discover(self, *module_paths: str):
        for module_path in module_paths:
            import_module(module_path)

    def provider(
        self,
        func: Callable = None,
        scope: str = None,
        name: str = None,
        lazy: bool = False,
    ) -> Provider:
        if func is None:
            return partial(self.provider, scope=scope, name=name, lazy=lazy)

        if scope is None:
            scope = self.default_scope
        else:
            scope = self.scope_aliases.get(scope, scope)

        if scope not in scopes.ALL:
            raise UnknownScope(scope)

        if name is None:
            name = func.__name__

        # NOTE: save the new provider before checking for recursion,
        # so that its dependants can detect it as a dependency.
        prov = Provider.create(func, name=name, scope=scope, lazy=lazy)
        self._add(prov)

        self._check_for_recursive_providers(name, func)

        return prov

    def _add(self, prov: Provider):
        self.providers[prov.name] = prov
        if isinstance(prov, SessionProvider):
            self._session_providers[prov.name] = prov

    def _check_for_recursive_providers(self, name: str, func: Callable):
        for other_name, other in self._get_providers(func).items():
            if name in self._get_providers(other.func):
                raise RecursiveProviderError(name, other_name)

    def _get_providers(self, func: Callable) -> Dict[str, Provider]:
        providers = {
            param: self._get(param)
            for param in inspect.signature(func).parameters
        }
        return dict(filter(lambda item: item[1] is not None, providers.items()))

    def _resolve_providers(self, consumer: Callable) -> ResolvedProviders:
        positional: PositionalProviders = []
        keyword: KeywordProviders = {}

        # NOTE: This flag goes down when we process a non-provider parameter.
        # It allows to detect provider parameters declared *after*
        # non-provider parameters.
        # processing_providers = True

        for name, parameter in inspect.signature(consumer).parameters.items():
            prov: Optional[Provider] = self.providers.get(name)

            if prov is None:
                positional.append((name, _NO_PROVIDER))
                continue

            if parameter.kind == inspect.Parameter.KEYWORD_ONLY:
                keyword[name] = prov
            else:
                positional.append((name, prov))

        return ResolvedProviders(positional=positional, keyword=keyword)

    def consumer(
        self, consumer: Union[partial, Callable, CoroutineFunction]
    ) -> CoroutineFunction:
        if isinstance(consumer, partial):
            if not inspect.iscoroutinefunction(consumer.func):
                raise ConsumerDeclarationError(
                    "'partial' consumers must wrap an async function"
                )
        elif not inspect.iscoroutinefunction(consumer):
            consumer = wrap_async(consumer)

        assert (
            isinstance(consumer, partial)
            and inspect.iscoroutinefunction(consumer.func)
            or inspect.iscoroutinefunction(consumer)
        )

        providers = self._resolve_providers(consumer)

        if not providers:
            return consumer

        @wraps(consumer)
        async def with_providers(*args, **kwargs):
            async with AsyncExitStack() as stack:

                async def _get_value(prov: Provider):
                    if prov.lazy:
                        return prov(stack)
                    return await prov(stack)

                args = list(args)
                injected_args = []
                for name, prov in providers.positional:
                    if prov is _NO_PROVIDER:
                        # No provider exists for this argument. Get it from
                        # `kwargs` in priority, and default to `args`.
                        if name in kwargs:
                            injected_args.append(kwargs.pop(name))
                            continue
                        with suppress(IndexError):
                            injected_args.append(args.pop())
                        continue
                    # A provider exists for this argument. Use it!
                    injected_args.append(await _get_value(prov))

                injected_kwargs = {
                    name: await _get_value(prov)
                    for name, prov in providers.keyword.items()
                    if name in kwargs
                }

                return await consumer(*injected_args, **injected_kwargs)

        return with_providers

    def freeze(self):
        """Resolve providers consumed by each provider."""
        for prov in self.providers.values():
            prov.func = self.consumer(prov.func)

    @contextmanager
    def exit_freeze(self):
        yield
        self.freeze()

    async def enter_session(self):
        for provider in self._session_providers.values():
            await provider.enter_session()

    async def exit_session(self):
        for provider in self._session_providers.values():
            await provider.exit_session()

    def session(self):
        return _Session(self)


# NOTE: can't use @asynccontextmanager from contextlib because it was
# only added in Python 3.7.
class _Session:
    def __init__(self, store: Store):
        self._store = store

    async def __aenter__(self):
        await self._store.enter_session()
        return None

    async def __aexit__(self, *args):
        await self._store.exit_session()
